@@ -1,20 +1,15 @@
-use std::{
-    collections::{HashMap, VecDeque},
-    net::SocketAddr,
-    time::Duration,
-};
+use std::{collections::HashMap, net::SocketAddr, time::Duration};
 
 use crate::{
     crypto::generate_random_bytes,
-    packet::{ConnectionKeepAlive, ConnectionRequest, EncryptedChallengeToken, Packet},
+    packet::{ChallengeToken, Packet},
     replay_protection::ReplayProtection,
     token::PrivateConnectToken,
-    NetcodeError, NETCODE_CONNECT_TOKEN_PRIVATE_BYTES, NETCODE_KEY_BYTES, NETCODE_MAC_BYTES, NETCODE_MAX_CLIENTS, NETCODE_MAX_PACKET_BYTES,
-    NETCODE_MAX_PAYLOAD_BYTES, NETCODE_MAX_PENDING_CLIENTS, NETCODE_SEND_RATE, NETCODE_USER_DATA_BYTES, NETCODE_VERSION_INFO,
+    ClientID, NetcodeError, PacketToSend, NETCODE_CONNECT_TOKEN_PRIVATE_BYTES, NETCODE_CONNECT_TOKEN_XNONCE_BYTES, NETCODE_KEY_BYTES,
+    NETCODE_MAC_BYTES, NETCODE_MAX_CLIENTS, NETCODE_MAX_PACKET_BYTES, NETCODE_MAX_PAYLOAD_BYTES, NETCODE_MAX_PENDING_CLIENTS,
+    NETCODE_SEND_RATE, NETCODE_USER_DATA_BYTES, NETCODE_VERSION_INFO,
 };
 
-type ClientID = u64;
-
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 enum ConnectionState {
     Disconnected,
@@ -46,20 +41,10 @@ struct ConnectTokenEntry {
     mac: [u8; NETCODE_MAC_BYTES],
 }
 
-/// Disconnect or connect event from the server
-#[derive(Debug, PartialEq, Eq)]
-pub enum ServerEvent {
-    /// A new connection has been stablished with a client.
-    /// This event contains the id from the connected client and the user data
-    /// that came from the connect token.
-    ClientConnected(ClientID, [u8; NETCODE_USER_DATA_BYTES]),
-    /// The client connection has been terminated.
-    ClientDisconnected(ClientID),
-}
-
 /// A server that can generate packets from connect clients, that are encrypted, or process
 /// incoming encrypted packets from clients. The server is agnostic from the transport layer, only
 /// consuming and generating bytes that can be transported in any way desired.
+#[derive(Debug)]
 pub struct NetcodeServer {
     clients: Box<[Option<Connection>]>,
     pending_clients: HashMap<SocketAddr, Connection>,
@@ -72,19 +57,24 @@ pub struct NetcodeServer {
     address: SocketAddr,
     current_time: Duration,
     global_sequence: u64,
-    events: VecDeque<ServerEvent>,
     out: [u8; NETCODE_MAX_PACKET_BYTES],
 }
 
 /// Result from processing an packet in the server
 #[derive(Debug, PartialEq, Eq)]
+#[allow(clippy::large_enum_variant)] // TODO: Consider boxing types
 pub enum ServerResult<'a, 's> {
     /// Nothing needs to be done.
     None,
     /// A packet to be sent back to the processed address.
-    PacketToSend(&'s mut [u8]),
+    PacketToSend(PacketToSend<'s>),
     /// A payload received from the client.
     Payload(ClientID, &'a [u8]),
+    /// A new client has connected
+    ClientConnected(ClientID, [u8; NETCODE_USER_DATA_BYTES], PacketToSend<'s>),
+    /// The client connection has been terminated.
+    // TODO: should we return the user_data also here?
+    ClientDisconnected(ClientID, Option<PacketToSend<'s>>),
 }
 
 impl NetcodeServer {
@@ -115,7 +105,6 @@ impl NetcodeServer {
             challenge_key,
             address,
             current_time,
-            events: VecDeque::new(),
             out: [0u8; NETCODE_MAX_PACKET_BYTES],
         }
     }
@@ -125,9 +114,8 @@ impl NetcodeServer {
         Self::new(Duration::ZERO, 32, 0, "127.0.0.1:0".parse().unwrap(), [0u8; NETCODE_KEY_BYTES])
     }
 
-    /// Returns an event from the server
-    pub fn get_event(&mut self) -> Option<ServerEvent> {
-        self.events.pop_front()
+    pub fn address(&self) -> SocketAddr {
+        self.address
     }
 
     fn find_or_add_connect_token_entry(&mut self, new_entry: ConnectTokenEntry) -> bool {
@@ -176,9 +164,30 @@ impl NetcodeServer {
     fn handle_connection_request<'a, 's>(
         &'s mut self,
         addr: SocketAddr,
-        request: &ConnectionRequest,
+        version_info: [u8; 13],
+        protocol_id: u64,
+        expire_timestamp: u64,
+        xnonce: [u8; NETCODE_CONNECT_TOKEN_XNONCE_BYTES],
+        data: [u8; NETCODE_CONNECT_TOKEN_PRIVATE_BYTES],
     ) -> Result<ServerResult<'a, 's>, NetcodeError> {
-        let connect_token = self.validate_client_token(request)?;
+        if version_info != *NETCODE_VERSION_INFO {
+            return Err(NetcodeError::InvalidVersion);
+        }
+
+        if protocol_id != self.protocol_id {
+            return Err(NetcodeError::InvalidProtocolID);
+        }
+
+        if self.current_time.as_secs() >= expire_timestamp {
+            return Err(NetcodeError::Expired);
+        }
+
+        let connect_token = PrivateConnectToken::decode(&data, self.protocol_id, expire_timestamp, &xnonce, &self.connect_key)?;
+
+        let in_host_list = connect_token.server_addresses.iter().any(|host| *host == Some(self.address));
+        if !in_host_list {
+            return Err(NetcodeError::NotInHostList);
+        }
 
         let addr_already_connected = find_client_mut_by_addr(&mut self.clients, addr).is_some();
         let id_already_connected = find_client_mut_by_id(&mut self.clients, connect_token.client_id).is_some();
@@ -193,7 +202,7 @@ impl NetcodeServer {
         }
 
         let mut mac = [0u8; NETCODE_MAC_BYTES];
-        mac.copy_from_slice(&request.data[NETCODE_CONNECT_TOKEN_PRIVATE_BYTES - NETCODE_MAC_BYTES..]);
+        mac.copy_from_slice(&data[NETCODE_CONNECT_TOKEN_PRIVATE_BYTES - NETCODE_MAC_BYTES..]);
         let connect_token_entry = ConnectTokenEntry {
             address: addr,
             time: self.current_time,
@@ -214,16 +223,17 @@ impl NetcodeServer {
                 Some((self.global_sequence, &connect_token.server_to_client_key)),
             )?;
             self.global_sequence += 1;
-            return Ok(ServerResult::PacketToSend(&mut self.out[..len]));
+            let packet_to_send = PacketToSend::new(addr, &mut self.out[..len]);
+            return Ok(ServerResult::PacketToSend(packet_to_send));
         }
 
         self.challenge_sequence += 1;
-        let packet = Packet::Challenge(EncryptedChallengeToken::generate(
+        let packet = Packet::generate_challenge(
             connect_token.client_id,
             &connect_token.user_data,
             self.challenge_sequence,
             &self.challenge_key,
-        )?);
+        )?;
 
         let len = packet.encode(
             &mut self.out,
@@ -243,51 +253,19 @@ impl NetcodeServer {
             send_key: connect_token.server_to_client_key,
             receive_key: connect_token.client_to_server_key,
             timeout_seconds: connect_token.timeout_seconds,
-            expire_timestamp: request.expire_timestamp,
+            expire_timestamp,
             user_data: connect_token.user_data,
             replay_protection: ReplayProtection::new(),
         });
         pending.last_packet_received_time = self.current_time;
         pending.last_packet_send_time = self.current_time;
 
-        Ok(ServerResult::PacketToSend(&mut self.out[..len]))
-    }
-
-    fn validate_client_token(&self, request: &ConnectionRequest) -> Result<PrivateConnectToken, NetcodeError> {
-        if request.version_info != *NETCODE_VERSION_INFO {
-            return Err(NetcodeError::InvalidVersion);
-        }
-
-        if request.protocol_id != self.protocol_id {
-            return Err(NetcodeError::InvalidProtocolID);
-        }
-
-        if self.current_time.as_secs() >= request.expire_timestamp {
-            return Err(NetcodeError::Expired);
-        }
-
-        let token = PrivateConnectToken::decode(
-            &request.data,
-            self.protocol_id,
-            request.expire_timestamp,
-            &request.xnonce,
-            &self.connect_key,
-        )?;
-
-        let in_host_list = token.server_addresses.iter().any(|host| *host == Some(self.address));
-        if in_host_list {
-            Ok(token)
-        } else {
-            Err(NetcodeError::NotInHostList)
-        }
+        let packet_to_send = PacketToSend::new(addr, &mut self.out[..len]);
+        Ok(ServerResult::PacketToSend(packet_to_send))
     }
 
     /// Returns an encoded packet payload to be sent to the client
-    pub fn generate_payload_packet<'s>(
-        &'s mut self,
-        client_id: ClientID,
-        payload: &[u8],
-    ) -> Result<(&'s mut [u8], SocketAddr), NetcodeError> {
+    pub fn generate_payload_packet<'s>(&'s mut self, client_id: ClientID, payload: &[u8]) -> Result<PacketToSend, NetcodeError> {
         if payload.len() > NETCODE_MAX_PAYLOAD_BYTES {
             return Err(NetcodeError::PayloadAboveLimit);
         }
@@ -296,7 +274,8 @@ impl NetcodeServer {
             let packet = Packet::Payload(payload);
             let len = packet.encode(&mut self.out, self.protocol_id, Some((client.sequence, &client.send_key)))?;
             client.sequence += 1;
-            return Ok((&mut self.out[..len], client.addr));
+
+            return Ok(PacketToSend::new(client.addr, &mut self.out[..len]));
         }
 
         Err(NetcodeError::ClientNotFound)
@@ -306,14 +285,17 @@ impl NetcodeServer {
     /// [ServerResult].
     pub fn process_packet<'a, 's>(&'s mut self, addr: SocketAddr, buffer: &'a mut [u8]) -> ServerResult<'a, 's> {
         match self.process_packet_internal(addr, buffer) {
-            Err(_) => ServerResult::None,
+            Err(_) => {
+                // TODO(log): error
+                ServerResult::None
+            }
             Ok(r) => r,
         }
     }
 
     fn process_packet_internal<'a, 's>(&'s mut self, addr: SocketAddr, buffer: &'a mut [u8]) -> Result<ServerResult<'a, 's>, NetcodeError> {
-        if buffer.len() <= 2 + NETCODE_MAC_BYTES {
-            return Ok(ServerResult::None);
+        if buffer.len() < 2 + NETCODE_MAC_BYTES {
+            return Err(NetcodeError::PacketTooSmall);
         }
 
         // Handle connected client
@@ -329,10 +311,10 @@ impl NetcodeServer {
                 ConnectionState::Connected => match packet {
                     Packet::Disconnect => {
                         client.state = ConnectionState::Disconnected;
-                        self.events.push_back(ServerEvent::ClientDisconnected(client.client_id));
+                        let client_id = client.client_id;
                         self.clients[slot] = None;
                         // TODO(log): debug
-                        return Ok(ServerResult::None);
+                        return Ok(ServerResult::ClientDisconnected(client_id, None));
                     }
                     Packet::Payload(payload) => {
                         if !client.confirmed {
@@ -341,7 +323,7 @@ impl NetcodeServer {
                         }
                         return Ok(ServerResult::Payload(client.client_id, payload));
                     }
-                    Packet::KeepAlive(_) => {
+                    Packet::KeepAlive { .. } => {
                         if !client.confirmed {
                             // TODO(log): debug
                             client.confirmed = true;
@@ -364,9 +346,20 @@ impl NetcodeServer {
             )?;
             pending.last_packet_received_time = self.current_time;
             match packet {
-                Packet::ConnectionRequest(request) => return self.handle_connection_request(addr, &request),
-                Packet::Response(response) => {
-                    let challenge_token = response.decode(&self.challenge_key)?;
+                Packet::ConnectionRequest {
+                    protocol_id,
+                    expire_timestamp,
+                    data,
+                    xnonce,
+                    version_info,
+                } => {
+                    return self.handle_connection_request(addr, version_info, protocol_id, expire_timestamp, xnonce, data);
+                }
+                Packet::Response {
+                    token_data,
+                    token_sequence,
+                } => {
+                    let challenge_token = ChallengeToken::decode(token_data, token_sequence, &self.challenge_key)?;
                     let mut pending = self.pending_clients.remove(&addr).unwrap();
                     if find_client_slot_by_id(&self.clients, challenge_token.client_id).is_some() {
                         // TODO(log): debug
@@ -379,23 +372,26 @@ impl NetcodeServer {
                             pending.state = ConnectionState::Disconnected;
                             self.global_sequence += 1;
                             pending.last_packet_send_time = self.current_time;
-                            return Ok(ServerResult::PacketToSend(&mut self.out[..len]));
+                            let packet_to_send = PacketToSend::new(addr, &mut self.out[..len]);
+                            return Ok(ServerResult::PacketToSend(packet_to_send));
                         }
                         Some(client_index) => {
-                            self.events
-                                .push_back(ServerEvent::ClientConnected(pending.client_id, pending.user_data));
                             let send_key = pending.send_key;
                             pending.state = ConnectionState::Connected;
                             pending.user_data = challenge_token.user_data;
                             pending.last_packet_send_time = self.current_time;
+                            let client_id: ClientID = pending.client_id;
+                            let user_data: [u8; NETCODE_USER_DATA_BYTES] = pending.user_data;
                             self.clients[client_index] = Some(pending);
-                            let packet = Packet::KeepAlive(ConnectionKeepAlive {
+
+                            let packet = Packet::KeepAlive {
                                 max_clients: self.max_clients as u32,
                                 client_index: client_index as u32,
-                            });
+                            };
                             let len = packet.encode(&mut self.out, self.protocol_id, Some((self.global_sequence, &send_key)))?;
                             self.global_sequence += 1;
-                            return Ok(ServerResult::PacketToSend(&mut self.out[..len]));
+                            let packet_to_send = PacketToSend::new(addr, &mut self.out[..len]);
+                            return Ok(ServerResult::ClientConnected(client_id, user_data, packet_to_send));
                         }
                     }
                 }
@@ -406,8 +402,14 @@ impl NetcodeServer {
         // Handle new client
         let (_, packet) = Packet::decode(buffer, self.protocol_id, None, None)?;
         match packet {
-            Packet::ConnectionRequest(request) => self.handle_connection_request(addr, &request),
-            _ => Ok(ServerResult::None), // Decoding packet without key can only return ConnectionRequest
+            Packet::ConnectionRequest {
+                data,
+                protocol_id,
+                expire_timestamp,
+                xnonce,
+                version_info,
+            } => self.handle_connection_request(addr, version_info, protocol_id, expire_timestamp, xnonce, data),
+            _ => unreachable!("Decoding packet without key can only return ConnectionRequest packets"),
         }
     }
 
@@ -446,21 +448,23 @@ impl NetcodeServer {
         self.pending_clients.retain(|_, c| c.state != ConnectionState::Disconnected);
     }
 
-    /// Updates the client, may return the client address and a packet to be sent to the client.
+    /// Updates the client, returns a ServerResult.
     ///
     /// # Example
     /// ```
+    /// # use renetcode::{PacketToSend, ServerResult};
     /// # let mut server = renetcode::NetcodeServer::__test();
     /// for client_id in server.clients_id().into_iter() {
-    ///     if let Some((packet, addr)) = server.update_client(client_id) {
-    ///         send_to(packet, addr);
+    ///     match server.update_client(client_id) {
+    ///         ServerResult::PacketToSend(PacketToSend { packet, address }) => send_to(packet, address),
+    ///         _ => { /* ... */ }
     ///     }
     /// }
     /// # fn send_to(p: &[u8], addr: std::net::SocketAddr) {}
     /// ```
-    pub fn update_client(&mut self, client_id: ClientID) -> Option<(&mut [u8], SocketAddr)> {
+    pub fn update_client<'s>(&'s mut self, client_id: ClientID) -> ServerResult<'_, 's> {
         let slot = match find_client_slot_by_id(&self.clients, client_id) {
-            None => return None,
+            None => return ServerResult::None,
             Some(slot) => slot,
         };
 
@@ -473,35 +477,41 @@ impl NetcodeServer {
             }
 
             if client.state == ConnectionState::Disconnected {
-                self.events.push_back(ServerEvent::ClientDisconnected(client.client_id));
                 let packet = Packet::Disconnect;
                 let sequence = client.sequence;
                 let send_key = client.send_key;
                 let addr = client.addr;
                 self.clients[slot] = None;
+
+                // TODO(log): error
                 let len = match packet.encode(&mut self.out, self.protocol_id, Some((sequence, &send_key))) {
-                    Err(_) => return None,
+                    Err(_) => return ServerResult::ClientDisconnected(client_id, None),
                     Ok(len) => len,
                 };
-                return Some((&mut self.out[..len], addr));
+
+                let packet_to_send = PacketToSend::new(addr, &mut self.out[..len]);
+                return ServerResult::ClientDisconnected(client_id, Some(packet_to_send));
             }
 
             if client.last_packet_send_time + NETCODE_SEND_RATE <= self.current_time {
-                let packet = Packet::KeepAlive(ConnectionKeepAlive {
+                let packet = Packet::KeepAlive {
                     client_index: slot as u32,
                     max_clients: self.max_clients as u32,
-                });
+                };
+
+                // TODO(log): error
                 let len = match packet.encode(&mut self.out, self.protocol_id, Some((client.sequence, &client.send_key))) {
-                    Err(_) => return None,
+                    Err(_) => return ServerResult::None,
                     Ok(len) => len,
                 };
                 client.sequence += 1;
                 client.last_packet_send_time = self.current_time;
-                return Some((&mut self.out[..len], client.addr));
+                let packet_to_send = PacketToSend::new(client.addr, &mut self.out[..len]);
+                return ServerResult::PacketToSend(packet_to_send);
             }
         }
 
-        None
+        ServerResult::None
     }
 
     pub fn is_client_connected(&self, client_id: ClientID) -> bool {
@@ -509,16 +519,24 @@ impl NetcodeServer {
     }
 
     /// Disconnect an client and returns its address and a disconnect packet to be sent to them.
-    pub fn disconnect(&mut self, client_id: ClientID) -> Result<(&mut [u8], SocketAddr), NetcodeError> {
+    // TODO: we can return Result<PacketToSend, NetcodeError>
+    //       but the library user would need to be aware that he has to run
+    //       the same code as Result::ClientDisconnected
+    pub fn disconnect<'s>(&'s mut self, client_id: ClientID) -> ServerResult<'_, 's> {
         if let Some(slot) = find_client_slot_by_id(&self.clients, client_id) {
             let client = self.clients[slot].take().unwrap();
-            self.events.push_back(ServerEvent::ClientDisconnected(client_id));
             let packet = Packet::Disconnect;
-            let len = packet.encode(&mut self.out, self.protocol_id, Some((client.sequence, &client.send_key)))?;
-            return Ok((&mut self.out[..len], client.addr));
+
+            // TODO(log): error
+            let len = match packet.encode(&mut self.out, self.protocol_id, Some((client.sequence, &client.send_key))) {
+                Err(_) => return ServerResult::ClientDisconnected(client_id, None),
+                Ok(len) => len,
+            };
+            let packet_to_send = PacketToSend::new(client.addr, &mut self.out[..len]);
+            return ServerResult::ClientDisconnected(client_id, Some(packet_to_send));
         }
 
-        Err(NetcodeError::ClientNotFound)
+        ServerResult::None
     }
 }
 
@@ -575,46 +593,52 @@ mod tests {
             private_key,
         )
         .unwrap();
-        let mut client = NetcodeClient::new(Duration::ZERO, connect_token);
+        let mut client = NetcodeClient::new(Duration::ZERO, client_id, connect_token);
         let (client_packet, _) = client.update(Duration::ZERO).unwrap();
 
         let result = server.process_packet(client_addr, client_packet);
         assert!(matches!(result, ServerResult::PacketToSend(_)));
         match result {
-            ServerResult::PacketToSend(packet) => client.process_packet(packet),
+            ServerResult::PacketToSend(PacketToSend { packet, .. }) => client.process_packet(packet),
             _ => unreachable!(),
         };
 
         assert!(!client.connected());
         let (client_packet, _) = client.update(Duration::ZERO).unwrap();
         let result = server.process_packet(client_addr, client_packet);
-        assert!(matches!(result, ServerResult::PacketToSend(_)));
 
         match result {
-            ServerResult::PacketToSend(packet) => client.process_packet(packet),
+            ServerResult::ClientConnected(r_id, r_data, PacketToSend { packet, .. }) => {
+                assert_eq!(client_id, r_id);
+                assert_eq!(user_data, r_data);
+                client.process_packet(packet)
+            }
             _ => unreachable!(),
         };
 
-        let client_connected = server.get_event().unwrap();
-        assert_eq!(client_connected, ServerEvent::ClientConnected(client_id, user_data));
-
         assert!(client.connected());
 
         let payload = [7u8; 300];
-        let (result, _) = server.generate_payload_packet(client_id, &payload).unwrap();
-        let result_payload = client.process_packet(result).unwrap();
+        let PacketToSend { packet, .. } = server.generate_payload_packet(client_id, &payload).unwrap();
+        let result_payload = client.process_packet(packet).unwrap();
         assert_eq!(payload, result_payload);
 
-        assert!(server.update_client(client_id).is_none());
+        let result = server.update_client(client_id);
+        assert_eq!(result, ServerResult::None);
         server.update(NETCODE_SEND_RATE);
 
-        let (keep_alive_packet, _) = server.update_client(client_id).unwrap();
-        assert!(client.process_packet(keep_alive_packet).is_none());
+        let result = server.update_client(client_id);
+        match result {
+            ServerResult::PacketToSend(PacketToSend { packet, .. }) => {
+                assert!(client.process_packet(packet).is_none());
+            }
+            _ => unreachable!(),
+        }
 
         let client_payload = [2u8; 300];
-        let (result, _) = client.generate_payload_packet(&client_payload).unwrap();
+        let PacketToSend { packet, .. } = client.generate_payload_packet(&client_payload).unwrap();
 
-        match server.process_packet(client_addr, result) {
+        match server.process_packet(client_addr, packet) {
             ServerResult::Payload(id, payload) => {
                 assert_eq!(id, client_id);
                 assert_eq!(client_payload, payload);
@@ -623,16 +647,17 @@ mod tests {
         }
 
         assert!(server.is_client_connected(client_id));
-        let (disconnect_packet, _) = server.disconnect(client_id).unwrap();
+        let result = server.disconnect(client_id);
+        match result {
+            ServerResult::ClientDisconnected(_client_id, Some(PacketToSend { packet, .. })) => {
+                assert!(client.connected());
+                assert!(client.process_packet(packet).is_none());
+                assert!(!client.connected());
+            }
+            _ => unreachable!(),
+        }
 
-        assert!(client.connected());
-        assert!(client.process_packet(disconnect_packet).is_none());
-
-        assert!(!client.connected());
         assert!(!server.is_client_connected(client_id));
-
-        let client_disconnected = server.get_event().unwrap();
-        assert_eq!(client_disconnected, ServerEvent::ClientDisconnected(client_id));
     }
 
     #[test]
